// BASED OF HOPEFRF SAMPLE CODE & http://dlehard.narod.ru/quick_start.pdf

#define F_CPU 8000000UL
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <util/delay.h>
#include "lcd_lib.h"
#include "uart.h"


/* RFM12B INTERFACE */
#define SCK 7 // SPI clock
#define MISO 6 // SPI Data input (RFM12B side)
#define MOSI 5 // SPI Data output (RFM12B side)
#define CS 3 // SPI SS (chip select) (PORTC)
#define NIRQ 4 // (PORTB)
/* IO CONTROL */
#define HI(x) PORTB |= (1<<(x))
#define LO(x) PORTB &= ~(1<<(x))
#define CS_HI() PORTC |= (1<<CS)
#define CS_LO() PORTC &= ~(1<<CS)
#define NIRQ_HI() PORTB |= (1<<NIRQ)
#define NIRQ_LO() PORTB &= ~(1<<NIRQ)
#define WAIT_NIRQ_LOW() while(PINB&(1<<NIRQ))

/* LED */
#define LED 1
#define LED_OFF() PORTA &= ~(1<<LED)
#define LED_ON() PORTA |= (1<<LED)


char display, distance10, distance100; //integer decimals of distance
int volatile pedal_period; //time between revs
int volatile pedal_period_ovlf; 
int volatile tire_period;
int volatile tire_period_ovlf;
unsigned volatile int revs;
unsigned volatile char revs_overflow;
unsigned volatile char pedalReset_High, wheelReset_High; //Flags to reset display on sensor timeout
char firstTime; //First time we update display

float volatile speed, distance, cadence, power, torque; //all our measurement values

int8_t lcd_buffer1[17];	// LCD display buffer
int8_t lcd_buffer2[17];	// LCD display buffer
int8_t lcd_buffer3[17];	// LCD display buffer
int8_t lcd_buffer4[17];	// LCD display buffer
int8_t lcd_buffer5[6];	// LCD display buffer
int8_t lcd_buffer6[6];	// LCD display buffer
int8_t lcd_buffer7[9];	// LCD display buffer
int8_t lcd_buffer8[9];	// LCD display buffer

char pedal_flag, tire_flag;

char volatile recieved[4];
char volatile recieveCount;
char volatile pedal_overflow_flag;
char volatile tire_overflow_flag;
float strain;
unsigned long tempLong;

float circ = 0.00128851;

union {
	float f;
	unsigned char b[4];
} u;

// LCD

const int8_t LCD_initialize[] PROGMEM = "LCD Initialized\0";
int8_t lcd_buffer[17];	// LCD display buffer

FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

void init_lcd(void)
{
	LCDinit();	//initialize the display
	LCDcursorOFF();
	LCDclr();				//clear the display
	LCDGotoXY(0,0);
	CopyStringtoLCD(LCD_initialize, 0, 0);
}

void portInit() {
	DDRB = (1<<MOSI) | (1<<SCK) | (1<<4);
	DDRA |= (1<<LED);
	DDRC |= (1<<CS);
	PORTB |= (1<<MISO); //Set MISO and SS to pulled up
	CS_HI();
}


unsigned char SPI_MasterTransmit(char cData)
{
	/* Start transmission */
	SPDR = cData;
	/* Wait for transmission complete */
	while
	(!(SPSR & (1<<SPIF)));
	return SPDR;
}


unsigned int writeCmd(unsigned int cmd) {
	unsigned int rec;
	unsigned char temp;
	CS_LO();
	temp = SPI_MasterTransmit((cmd >> 8));
	rec = SPI_MasterTransmit(cmd);
	CS_HI();
	return (rec | (temp << 8));
}

unsigned char rfRecv() {
	unsigned int data;
	data = writeCmd(0xB000);
	return (data&0x00FF);
}

//Clear rf module fifo
void FIFOReset() {
	writeCmd(0xCA81);
	writeCmd(0xCA83);
}

void spiInit() {
	SPCR = (1<<SPE) | (1<<MSTR);
	// no interupt | enable SPI | Master | CPOL=0 | CPHA=0 | Fosc/4 |
}


void rfInit() {

	//Set up INT for reading (INT2)
	PCICR = (1<<PCIE1);
	PCMSK1 = (1<<PCINT12);
	//Set up rf module
	writeCmd(0x80F7); //EL,EF,413band,12.0pF
	writeCmd(0x8299); //er,!ebb,ET,ES,EX,!eb,!ew,DC
	writeCmd(0xA640); //frequency select
	writeCmd(0xC647); //4.8kbps
	writeCmd(0x94C0); //VDI,FAST,134kHz,0dBm,-103dBm
	writeCmd(0xC2AC); //AL,!ml,DIG,DQD4
	writeCmd(0xCA11); //FIFO1,SYNC,!ff,DR
	writeCmd(0xCED4); //SYNC=2DD4,AG
	writeCmd(0xC483); //@PWR,NO RSTRIC,!st,!fi,OE,EN
	writeCmd(0x9820); //!mp,45kHz,MAX OUT
	writeCmd(0xCCA7); //OB1,ACOB0, LPX,Iddy,CDDIT,CBW0
	writeCmd(0xE000); //NOT USED
	writeCmd(0xC800); //NOT USED
	writeCmd(0xC040); //1.66MHz,2.2V
}

void timerInit()
{
	TCCR2B = 7; //Prescaler to 1024
	TIMSK2 = 1 ;//Overflow interrupt enable

	TCCR0B = 5; //Prescaler to 1024
	TIMSK0 = 1;//Overflow interrupt enable

	TCCR1B = 4; // /256
	TIMSK1 = 1; //Overflow interrupt enable

}

void interruptInit()
{
	EIMSK = (1<<INT0) | (1<<INT1) | (1<<INT2); //Turn on external
	EICRA = (1<<ISC01) | (1<<ISC11) | (1<<ISC20) | (1<<ISC21); //Falling edge for 0/1 - Low for 2

}

//Pedal hall effect sensor interrupt routine
ISR(INT0_vect)
{
	pedal_period = TCNT0 + pedal_period_ovlf  ;
    TCNT0 = 0 ;
    pedal_period_ovlf = 0;
	pedal_flag = 1;
}

//Tire hall effect sensor interrupt routine
ISR(INT1_vect)
{
	if (revs < 65535) revs++;
	else
	{
		revs = 0;
		revs_overflow++;
	}
	tire_period = TCNT2 + tire_period_ovlf  ;
    TCNT2 = 0 ;
    tire_period_ovlf = 0 ;
	tire_flag = 1;
}

//Fifo half full interrupt generated by rf module
ISR(INT2_vect)
{
	u.b[recieveCount] = rfRecv();
	recieveCount++;
	if (recieveCount == 4)
	{
		torque = u.f;
		FIFOReset();
	}
}

//Timer 0 overflow, 
ISR (TIMER0_OVF_vect) {
        pedal_period_ovlf = pedal_period_ovlf + 256 ;
		if (pedal_period_ovlf > 34) pedal_overflow_flag = 1;
}

//Timer 1 overflow, no new movement detected so reset display values
ISR (TIMER1_OVF_vect) {
		pedalReset_High++;
		wheelReset_High++;
		if (pedalReset_High >= 2) {
			pedalReset_High = 0;
			cadence = 0;
		}
		if (wheelReset_High >= 2) {
			wheelReset_High = 0;
			speed = 0;
		}
}

ISR (TIMER2_OVF_vect) {
        tire_period_ovlf = tire_period_ovlf + 256 ;
		if (pedal_period_ovlf > 34) tire_overflow_flag = 1;
}


int main(void) {
	unsigned char data, i;
	init_lcd();
	uart_init();
  	stdout = stdin = stderr = &uart_str;
  	fprintf(stdout,"Starting...\n\r");
	portInit();
	spiInit();
	rfInit();
	timerInit();
	interruptInit();
	FIFOReset();
	revs = 0;
	firstTime = 1;
	pedalReset_High = 0;
	wheelReset_High = 0;
	cadence = 0.0;
	speed= 0.0;
	sei();

	while(1) {
		//If we have received all 4 bytes of the float
		if(recieveCount == 4)
		{
			//If it is time to update the display
			if (display)
			{
				if (firstTime)
				{
					sprintf(lcd_buffer1,"Distance: %d.%d      ", (int)distance,(int)distance10);
					sprintf(lcd_buffer2,"Cadence: %u  " ,(int)cadence);
					sprintf(lcd_buffer3,"Speed: %u mph  ",(int)speed);
					sprintf(lcd_buffer4,"Power: %u W  ",(int)(cadence*torque*0.024));
					LCDGotoXY(0,0);
					LCDstring(lcd_buffer1, strlen(lcd_buffer1));
					LCDGotoXY(0,1);
					LCDstring(lcd_buffer2, strlen(lcd_buffer2));
					LCDGotoXY(-4,2);
					LCDstring(lcd_buffer3, strlen(lcd_buffer3));
					LCDGotoXY(-4,3);
					LCDstring(lcd_buffer4, strlen(lcd_buffer4));
					fprintf(stdout,"%.2f,%.2f,%.2f,%.2f\n\r", cadence, speed, (cadence*torque*0.024), torque); //For logging in terminal
				} else
				{
					sprintf(lcd_buffer5,"%d.%d  ", (int)distance, (int)distance10);
					sprintf(lcd_buffer6,"%u   " ,(int)cadence);
					sprintf(lcd_buffer7,"%u mph  ",(int)speed);
					sprintf(lcd_buffer8,"%u W  ",(int)(cadence*torque*0.024));
					LCDGotoXY(10,0);
					LCDstring(lcd_buffer5, strlen(lcd_buffer5));
					LCDGotoXY(9,1);
					LCDstring(lcd_buffer6, strlen(lcd_buffer6));
					LCDGotoXY(3,2);
					LCDstring(lcd_buffer7, strlen(lcd_buffer7));
					LCDGotoXY(3,3);
					LCDstring(lcd_buffer8, strlen(lcd_buffer8));
					fprintf(stdout,"%.2f,%.2f,%.2f,%.2f\n\r", cadence, speed, (cadence*torque*0.024), torque);//For logging in terminal
				}
			}
			display ^= 1; //Display every other new value
			recieveCount = 0;
		}

		//If there is a new pedal rotation
		if (pedal_flag)
		{
			cadence = (468750.0/pedal_period); //Calculate cadence in RPM
			if((cadence > 200) || (cadence < 0)){
				cadence = 0.0;
			}
			pedal_flag = 0;
			pedalReset_High = 0; //Kick counter back to 0
		}
		//If there is a new tire rotation
		if (tire_flag)
		{
			speed = (468750.0/tire_period)*6.8*0.01136; //Speed in MPH
			if((speed > 50.0) || (speed < 0)){
				speed = 0.0;
			}
			distance = (revs_overflow*65535 + revs)*circ;
			distance10 = (char)(distance*10)%10; //calcualted tenths for faster printing
			tire_flag = 0;
			wheelReset_High = 0; //Kick counter back to 0
		}

		//If the odometer reset button is pressed
		if (!(PINA & 1))
		{
			revs = 0;
			revs_overflow = 0;
		}

	}
}
